/* 
 * $RCSfile: process.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 *
 * $Log: process.h,v $
 * Revision 1.3  1996/11/04 22:35:44  mjk
 * *** empty log message ***
 *
 * Revision 1.2  1996/03/11 18:01:47  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/02/19  20:15:16  mjk
 * Initial revision
 *
 *
 */

#ifndef process_h
#define process_h

#include "thread.h"
#include "semaphore.h"
#include "xtype.h"

/* 
 * Priorities here (gotta check with pthread spec) */

#define	THREAD_STDPRIO	PTHREAD_DEFAULT_PRIORITY
#define THREAD_MAXPRIO	PTHREAD_MAX_PRIORITY
#define THREAD_MINPRIO	PTHREAD_MIN_PRIORITY

#ifndef NULL
#define NULL 0
#endif

typedef void (*CreateProcFunc) ( int, ... );

#define CreateProcess0 CreateProcess
#define CreateProcess1 CreateProcess
#define CreateProcess2 CreateProcess
#define CreateProcess3 CreateProcess
#define CreateProcess4 CreateProcess
#define CreateProcess5 CreateProcess
#define CreateProcess6 CreateProcess
#define xCreateProcess CreateProcess

typedef struct _Process {
  struct _Process	*link;
  pthread_t		thread;
} Process;

#define STACKSIZE (32*1024)

typedef sema_t	Semaphore;	/* directly use the POSIX struct */

extern Process *Active;
extern int      SignalsPossible;

#define kSwitch() Yield()

/* The -DXKLOCKDEBUG switch provides for counting the locking depth, and
   complaining if it isn't 1.  It has a bug, in that the increment & decrement
   instructions aren't atomic in RISC architectures.  Presumably, the occasions
   where this fails are rare.  */

#ifdef XKLOCKDEBUG
extern int xklockdepthreq;
extern int xklockdepth;
extern int tracexklock;
#endif

/* the master concurrency locks */
extern pthread_mutex_t sledgehammer_concurrency_control;

#ifndef XKLOCKDEBUG

#define MASTER_LOCK pthread_mutex_lock(&sledgehammer_concurrency_control)
#define MASTER_UNLOCK pthread_mutex_unlock(&sledgehammer_concurrency_control)

#else	/* XKLOCKDEBUG */

#define MASTER_LOCK                                                           \
do {xklockdepthreq++;                                                         \
 xTrace1(xklock,TR_EVENTS,"requesting xklock, depthreq %d",xklockdepthreq);   \
 pthread_mutex_lock(&sledgehammer_concurrency_control);                       \
 xklockdepth++;                                                               \
 if (xklockdepth!=1)                                                          \
 { xTrace1(xklock,TR_ERRORS,"got xklock, wrong depth %d",xklockdepth); };     \
 xTrace2(xklock,TR_EVENTS,"got xklock, depth %d, depthreq %d",                \
         xklockdepth,xklockdepthreq); } while(0)
#define MASTER_UNLOCK                                                         \
do {if (xklockdepth!=1)                                                       \
 {xTrace1(xklock,TR_ERRORS,"giving up xklock, wrong depth %d",xklockdepth); };\
 xTrace2(xklock,TR_EVENTS,"giving up xklock, depth %d, depthreq %d",          \
         xklockdepth,xklockdepthreq);                                         \
 xklockdepth--;                                                               \
 pthread_mutex_unlock(&sledgehammer_concurrency_control);                     \
 xTrace1(xklock,TR_EVENTS,"gave up xklock, depthreq %d",xklockdepthreq);      \
 xklockdepthreq--; } while(0)

#endif XKLOCKDEBUG

#ifdef __STDC__


void    xk_master_lock( void );
void    xk_master_unlock( void );
void xk_pthread_sem_wait(Semaphore *s);
void xk_pthread_sem_signal(Semaphore *s);
void semWait(Semaphore *s);
void semSignal(Semaphore *s);
void semInit(Semaphore *s,unsigned n);
void Yield(void);

bool CreateProcess(void *, int, int, ... );
void threadInit( void );

#else

void    xk_master_lock( );
void    xk_master_unlock( );
void semWait( );
void semSignal( );
void semInit( );
void Yield( );

bool CreateProcess( );
void threadInit( );

#endif
#endif 
