/*
 * $RCSfile: lcp_options.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: lcp_options.c,v $
 * Revision 1.3  1994/10/28 21:02:35  andrey
 * *** empty log message ***
 *
 * Revision 1.2  1994/10/27  17:27:04  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:46:02  andrey
 * Initial revision
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/10/28 21:02:35 $
 */

/*
 * lcp option specific routines to use with ppp_options common
 *   list routines
 */

static void * lcp_duplicate_option(void * this, u_char type);
static void lcp_destroy_opt( void * this, u_char type);
static void lcp_print_opt(void * this, u_char type);
static int lcp_option_matches(void * from, void * to, u_char type);

static void *
lcp_duplicate_option(void * this, u_char type)
{
  void * that = 0;

  switch(type) {
    case LCP_MRU:
      that = X_NEW(MRU_opt);
      ((MRU_opt*)that)->mru = ((MRU_opt*)this)->mru;
    break;
    case LCP_AUTH:
      that = X_NEW(AUTH_opt);
      ((AUTH_opt*)that)->auth_prot = ((AUTH_opt*)this)->auth_prot;
    break;
    case LCP_QUAL:
      that = X_NEW(QUAL_opt);
      ((QUAL_opt*)that)->quality_prot = ((QUAL_opt*)this)->quality_prot;
    break;
    case LCP_PFC:
    case LCP_ACFC:
      that = 0;
    break;
    case LCP_MAGIC:
      that = X_NEW(MAGIC_opt);
      ((MAGIC_opt*)that)->magic_number = ((MAGIC_opt*)this)->magic_number;
    break;
    case LCP_ACCM:
      that = X_NEW(ACCM_opt);
      ((ACCM_opt*)that)->accm = ((ACCM_opt*)this)->accm;
    break;

    case LCP_UNKNOWN:
      that = X_NEW(UNKNOWN_opt);
      ((UNKNOWN_opt *)that)->type = ((UNKNOWN_opt *)this)->type;
      ((UNKNOWN_opt *)that)->length = ((UNKNOWN_opt *)this)->length;
      ((UNKNOWN_opt *)that)->buf = xMalloc(((UNKNOWN_opt *)this)->length);
      bcopy(((UNKNOWN_opt *)this)->buf, ((UNKNOWN_opt *)that)->buf,
            ((UNKNOWN_opt *)that)->length);
    break;
  }

  return that;
}

static void
lcp_destroy_opt(void * this, u_char type)
{
  switch(type) {
    case LCP_PFC:
    case LCP_ACFC:
    break;

    case LCP_UNKNOWN:
      FREE(((UNKNOWN_opt *)this)->buf);
    case LCP_MRU:
    case LCP_AUTH:
    case LCP_QUAL:
    case LCP_MAGIC:
    case LCP_ACCM:
      FREE(this);
    break;
  }
}

static void
lcp_print_opt(void * this, u_char type)
{
  switch(type) {
    case LCP_PFC:
      printf("pfc\t");
    break;
    case LCP_ACFC:
      printf("acfc\t");
    break;
    case LCP_UNKNOWN:
      printf("unknown\t");
    break;
    case LCP_MRU:
      printf("mru %d\t", ((MRU_opt*)this)->mru);
    break;
    case LCP_AUTH:
      printf("auth %x\t", ((AUTH_opt*)this)->auth_prot);
    break;
    case LCP_QUAL:
      printf("qual %x\t", ((QUAL_opt*)this)->quality_prot);
    break;
    case LCP_MAGIC:
      printf("magic %x\t", ((MAGIC_opt*)this)->magic_number);
    break;
    case LCP_ACCM:
      printf("accm %x\t", ((ACCM_opt*)this)->accm);
    break;
  }
}

static int
lcp_option_matches(void * from, void * to, u_char type)
{
  switch(type) {
    case LCP_MRU:
         return ((MRU_opt *)from)->mru == ((MRU_opt *)to)->mru;
    break;
    case LCP_ACCM:
         return ((ACCM_opt *)from)->accm == ((ACCM_opt *)to)->accm;
    break;
    case LCP_AUTH:
         return ((AUTH_opt *)from)->auth_prot == ((AUTH_opt *)to)->auth_prot;
    break;
    case LCP_QUAL:
         return ((QUAL_opt *)from)->quality_prot ==
                  ((QUAL_opt *)to)->quality_prot;
    break;
    case LCP_MAGIC:
         return ((MAGIC_opt *)from)->magic_number ==
                ((MAGIC_opt *)to)->magic_number;
    break;

    case LCP_PFC:
    case LCP_ACFC:
      return 1;
    break;
    default:
      return 0;
  }
}


