/*
 * photuris.c
 *
 * x-kernel v3.2
 *
 */


int tracephoturisp;

#include "xkernel.h"
#include "udp.h"
#include "md5hash.h"
#include "saidmgr.h"
#include "memory.h"
#include "crypt.h"
#include "km.h"
#include "dheckxleaf.h"
#include "cookie_photuris.h"
#include "photuris_i.h"
#include <stdlib.h>

int photurisp;

#ifdef __STDC__

void			photuris_init( XObj );
static void		photurisSessnInit( XObj );
static XObj		photurisOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	photurisOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t 	photurisOpenEnable( XObj, XObj, XObj, Part *);
static xkern_return_t	photurisPop( XObj, XObj, Msg *, void * );
static xkern_return_t	photurisDemux( XObj, XObj, Msg * );
static xkern_return_t	photurisClose( XObj );
static XObj		photurisCreateSessn( XObj, XObj, XObj );
static void 		photurisSessnInit(XObj);
static long 		simple_pop(void *, char *, long, void *);
static void 		simple_push(void *, char *, long, void * );
static long 		long_pop(void *, char *, long, void *);
static void 		long_push(void *, char *, long, void * );
static void 		encrypt_signature_message(Msg *, XObj);
static void 		decrypt_signature_message(Msg *, XObj);
static unsigned int 	get_size_of_variable_precision_number(Msg *);
static void 		get_attribute(Msg *, Attribute *);
static void 		get_attribute_list(XObj, Msg *, Attribute_list *);
static void 		get_cookie_response(XObj, Msg *, Cookie_Response *);
static void 		get_key_request(XObj, Msg *, Key_Request *);
static void 		get_key_response(XObj, Msg *, Key_Response *);
static void 		get_signature_message(XObj, Msg *, Signature_Message *);
static void 		get_key_change(XObj, Msg *, Key_Change *);
static void 		put_identity_verification(Msg * , Public_value *,Groupe_list *,
					Public_value *, 
					Public_value *, Attribute_list *,
					Attribute_list *, unsigned char *,
					int *,tSPI *, Certificate *, 
					Attribute_list *, 
					Key * );
static void	 	put_size_of_variable_precision_number(Msg *,unsigned int);
static void 		put_attribute(Msg *, Attribute *);
static void 		put_attribute_list(Msg *, Attribute_list *);
static void 		put_cookie_request(Msg *, Cookie_Request *);
static void 		put_cookie_response(Msg *, Cookie_Response *);
static void 		put_key_request(Msg *, Key_Request *);
static void 		put_key_response(Msg *, Key_Response *);
static void 		put_signature_message(Msg *, Signature_Message *);
static void 		put_key_change(Msg *, Key_Change *);
static int 		key_request_equal(Key_Request *, Key_Request *);
static int 		signature_message_equal(Signature_Message *, Signature_Message *);
static void 		calculate_session_key(XObj , Key * , Attribute *, Public_value *, 
					Signature *, Signature *, Cookie *, Cookie *, 
					tSPI *);		
static void 		calculate_raw_key(XObj);
static void 		set_session_key(XObj);
static	void 		calculate_signature(XObj, Signature *, Certificate *, unsigned char );
static	int 		check_signature(XObj, Signature *, Certificate *,unsigned char);
static void 		do_cookie_request(XObj, Msg *);
static void 		do_cookie_response(XObj, Msg *);
static void 		do_key_request(XObj, Msg *);
static void 		do_key_response(XObj, Msg *);
static void 		do_signature_message_initiator(XObj, Msg *);
static void 		do_signature_message_responder(XObj, Msg *);
static void 		do_received_signature_message_responseer(XObj, Msg *);
static void 		do_key_change(XObj, Msg *);
static void 		do_error_message(XObj, Msg *);
static void		generate_error_message(XObj, unsigned char);
static void 		timeout(Event, void*);
static int 		timeouts(void *, void *, void *);
#else
void			photuris_init( );
static void		photurisSessnInit();
static XObj		photurisOpen( );
static xkern_return_t	photurisOpenDone( );
static xkern_return_t 	photurisOpenEnable( );
static xkern_return_t	photurisPop( );
static xkern_return_t	photurisDemux( );
static xkern_return_t	photurisClose( );
static XObj		photurisCreateSessn( );
static void 		photurisSessnInit();
static long		simple_pop();
static void 		simple_push();
static long		long_pop();
static void 		long_push();
static void 		encrypt_signature_message();
static void 		decrypt_signature_message();
static unsigned int 	get_size_of_variable_precision_number();
static void 		get_attribute();
static void 		get_attribute_list();
static void 		get_cookie_response();
static void 		get_key_request();
static void 		get_key_response();
static void 		get_signature_message();
static void 		get_key_change();
static void 		put_identity_verification();
static void 		put_attribute();
static void 		put_attribute_list();
static void	 	put_size_of_variable_precision_number();
static void 		put_cookie_request();
static void 		put_cookie_response();
static void 		put_key_request();
static void 		put_key_response();
static void 		put_signature_message();
static void 		put_key_change();
static int 		key_request_equal();
static int 		signature_message_equal();
static void 		calculate_session_key();		
static void 		calculate_raw_key();
static void 		set_session_key();
static	void 		calculate_signature();
static	int 		check_signature();
static void 		do_cookie_request();
static void 		do_cookie_response();
static void 		do_key_request();
static void 		do_key_response();
static void 		do_signature_message_initiator();
static void 		do_signature_message_responder();
static void 		do_received_signature_message_responseer();
static void 		do_key_change();
static void 		do_error_message();
static void		generate_error_message();
static void 		timeout();
static int 		timeouts();

#endif

#define roundup8(x) (((x)/8)+(((x)%8)?1:0))

static int timeouts(key, value, arg)
void * key;
void * value;
void * arg;
{
	XObj	self = (XObj) value;
	SState * ss = ((SState *) self->state);
	int 	event;

	xTrace2(photurisp, TR_EVENTS, "timeout %08x %08x entered",*((XObj *)key),value);

	xDuplicate(self); /* make sure xPop doesn't close the session */
	if (ss->global_timeout == 0)	
	{
		event = GLOBAL_TIMEOUT;	
		xPop(self, xGetDown(self,0), NULL, &event);
	}
	else if (ss->timeout == 0)
	{
		event = TIMEOUT;	
		xPop(self, xGetDown(self,0), NULL, &event);
	}
	if ( ss->timeout >= 0 )
		ss->timeout --;
	if ( ss->global_timeout >= 0 )
		ss->global_timeout --;
	xClose(self);
	
	return MFE_CONTINUE;
}

static void timeout(ev, me)
Event ev; 
void * me;
{
	XObj	self = (XObj) me;
	PState * ps  = (PState *)self->state; 
	Map	actM =  ((PState *)self->state)->actMap;
	if (ps->group_timeout == 0)	
	{
		int x;
		int	random;
		random = rand();
		xTrace0(photurisp, TR_EVENTS, "recalculate groups");
		ps->group_timeout = ps->group_timeout_ini;
		xControl(xGetDown(self,0),COOKIE_PHOTURIS_SET_RANDOM_VALUE,(char *)&random,sizeof(int));
		for(x=0; x<ps->supported_groups.length; x++)
		{
			xClose(ps->group_sessions[x]);
			ps->group_sessions[x]=xOpen(self,self,ps->group_protocols[x],NULL);
		}	
		xTrace0(photurisp, TR_EVENTS, "recalculate groups finished");
	}
	if ( ps->group_timeout >= 0 )
		ps->group_timeout --;
	mapForEach(actM,timeouts,NULL);
	evDetach(((PState*)self->state)->timeout = evSchedule(timeout,self, 
				((PState*)self->state)->timeout_time));

}



static long simple_pop(dst, src, count, type)
    void * dst;
    char * src;
    long count;
    void * type;
{
	memcpy(dst,src,count);
	return count;
}

static void simple_push(src,dst,count,type)
    void * src;
    char * dst;
    long count;
    void * type;
{
	memcpy(dst,src,count);
}

static long long_pop(dst, src, count, type)
    void * dst;
    char * src;
    long count;
    void * type;
{
	unsigned long *	ds = (unsigned long *) dst;
	unsigned long  d =0;
	memcpy((((char *)&d)+(4-count)),src,count);
	*ds = ntohl (d);
	return count;
}

static void long_push(src,dst,count,type)
    void * src;
    char * dst;
    long count;
    void * type;
{
	unsigned long *	s = (unsigned long *) src;
	unsigned long d=0;
	d = htonl(*s);
	memcpy(dst,(((char *)&d)+(4-count)),count);
}

static void encrypt_signature_message(msg,self)
	Msg * msg;
	XObj self;
{
	XObj s;
	int x;
	XObj cp;
	XObj md5;
	Part p[3];
	PState * ps = ((PState*)self->myprotl->state);
	SState * ss = ((SState*)self->state);
	
	partInit(p,3);

	switch(ss->Group_index.index)
	{
	/* DES CBC 64 bit IV*/
	case 01:
	case 02:
		{
		unsigned char key[8];
		Msg  rmsg;
		msgConstructEmpty(&rmsg);
		for(x=0; (x < ps->supported_attributes.length)
			 && 18 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("crypt as Anonymity_Choice crypt is not configured but used");
			msgDestroy(&rmsg);
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		cp = ps->attribute_protocols[x];
		for(x=0; (x < ps->supported_attributes.length)
			 && 5 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("crypt as Anonymity_Choice md5 is not configured but used");
			msgDestroy(&rmsg);
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		
		md5 = ps->attribute_protocols[x];
	
		msgPush(&rmsg,simple_push,ss->gxy.value,roundup8(ss->gxy.length),NULL);
		msgPush(&rmsg,simple_push,ss->gy.value,roundup8(ss->gy.length),NULL);
		msgPush(&rmsg,simple_push,ss->gx.value,roundup8(ss->gx.length),NULL);
		msgPush(&rmsg,simple_push,ss->R_Cookie.value,COOKIE_SIZE,NULL);
		msgPush(&rmsg,simple_push,ss->I_Cookie.value,COOKIE_SIZE,NULL);
		msgPush(&rmsg,simple_push,ss->gxy.value,roundup8(ss->gxy.length),NULL);

		partPush(p[LOCAL_PART],(void *)&ss->myhost,sizeof(IPhost));
		partPush(p[SPI_PART],(void *)(ss->state > 128)?&ss->R_SPI:&ss->I_SPI,sizeof(tSPI));
		partPush(p[SPI_PART],(void *)&ss->myhost,sizeof(IPhost));
		if ((s = xOpen(self->myprotl,self->myprotl,md5,p))==ERR_XOBJ)
		{
			xError("crypt as Anonymity_Choice couldn't open md5 session");
			msgDestroy(&rmsg);
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		xPush(s,&rmsg);
		xClose(s);
		msgPop(&rmsg,simple_pop,(char*) key, 8, NULL);

		xIfTrace(photurisp, TR_DETAILED)
			printf("encrypt key %02x %02x %02x %02x %02x %02x %02x %02x  \n",
			key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7]); 

		partPush(p[LOCAL_PART],(void *)&ss->myhost,sizeof(IPhost));
		partPush(p[SPI_PART],(void *)(ss->state > 128)?&ss->R_SPI:&ss->I_SPI,sizeof(tSPI));
		partPush(p[SPI_PART],(void *)&ss->myhost,sizeof(IPhost));

		if ((s = xOpen(self->myprotl,self->myprotl,cp,p))==ERR_XOBJ)
		{
			xError("crypt as Anonymity_Choice couldn't open crypt session");
			msgDestroy(&rmsg);
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		xControl(s,CRYPT_OUT_SET,key,8);
		xPush(s,msg);
		xControl(s,CRYPT_OUT_INVALIDATE,NULL,0);
		xClose(s);
		msgDestroy(&rmsg);
		break;
		}
	default:
		xError("Unsupported Anonymity_Choice in encrypt_signature_message");
		ss->state = DELETED_STATE;
		generate_error_message(self,0);
	}
}

static void decrypt_signature_message(msg,self)
	Msg * msg;
	XObj self;
{
	char t = 04;
	XObj s;
	int x;
	XObj cp;
	XObj md5;
	Part p[3];
	PState * ps = ((PState*)self->myprotl->state);
	SState * ss = ((SState*)self->state);

	partInit(p,3);
        /* add message id which was striped in demux */
	msgPush(msg,simple_push,&t,sizeof(char),NULL);

	switch(ss->Group_index.index)
	{
	/* DES CBC */
	case 01:
	case 02:
		{
		unsigned char key[8];
		Msg  rmsg;
		msgConstructEmpty(&rmsg);
		for(x=0; (x < ps->supported_attributes.length)
			 && 18 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("crypt as Anonymity_Choice crypt is not configured but used");
			ss->state = DELETED_STATE;
			generate_error_message(self,0);
			msgDestroy(&rmsg);
			break;
		}
		cp = ps->attribute_protocols[x];
		for(x=0; (x < ps->supported_attributes.length)
			 && 5 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("crypt as Anonymity_Choice md5 is not configured but used");
			ss->state = DELETED_STATE;
			generate_error_message(self,0);
			msgDestroy(&rmsg);
			break;
		}
		md5 = ps->attribute_protocols[x];

		msgPush(&rmsg,simple_push,ss->gxy.value,roundup8(ss->gxy.length),NULL);
		msgPush(&rmsg,simple_push,ss->gx.value,roundup8(ss->gx.length),NULL);
		msgPush(&rmsg,simple_push,ss->gy.value,roundup8(ss->gy.length),NULL);
		msgPush(&rmsg,simple_push,ss->R_Cookie.value,COOKIE_SIZE,NULL);
		msgPush(&rmsg,simple_push,ss->I_Cookie.value,COOKIE_SIZE,NULL);
		msgPush(&rmsg,simple_push,ss->gxy.value,roundup8(ss->gxy.length),NULL);


		partPush(p[LOCAL_PART],(void *)&ss->myhost,sizeof(IPhost));
		partPush(p[SPI_PART],(void *)(ss->state > 128)?&ss->R_SPI:&ss->I_SPI,sizeof(tSPI));
		partPush(p[SPI_PART],(void *)&ss->myhost,sizeof(IPhost));
		if ((s = xOpen(self->myprotl,self->myprotl,md5,p))==ERR_XOBJ)
		{
			xError("crypt as Anonymity_Choice couldn't open md5 session");
			msgDestroy(&rmsg);
			ss->state = DELETED_STATE;
			generate_error_message(self,0);
			break;
		}
		xPush(s,&rmsg);
		xClose(s);
		msgPop(&rmsg,simple_pop,(char*) key, 8, NULL);

		xIfTrace(photurisp, TR_DETAILED)
			printf("decrypt key %02x %02x %02x %02x %02x %02x %02x %02x \n",
			key[0],key[1],key[2],key[3],key[4],key[5],key[6],key[7]); 

		partPush(p[LOCAL_PART],(void *)&ss->myhost,sizeof(IPhost));
		partPush(p[SPI_PART],(void *)(ss->state > 128)?&ss->R_SPI:&ss->I_SPI,sizeof(tSPI));
		partPush(p[SPI_PART],(void *)&ss->myhost,sizeof(IPhost));

		if ((s = xOpen(self->myprotl,self->myprotl,cp,p))==ERR_XOBJ)
		{
			xError("crypt as Anonymity_Choice couldn't open crypt session");
			msgDestroy(&rmsg);
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		xControl(s,CRYPT_OUT_SET,key,8);
		s->up=NULL;
		xPop(s,self,msg,NULL);
		xControl(s,CRYPT_OUT_INVALIDATE,NULL,0);
		xClose(s);
		break;
		msgDestroy(&rmsg);
		}
	default:
		xError("Unsupported Anonymity_Choice in decrypt_signature_message");
		ss->state = DELETED_STATE;
		generate_error_message(self,0);
	}
	/* strip message type up again */
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	
}



static unsigned int get_size_of_variable_precision_number(msg)
	Msg * msg;
{
	unsigned char t;
	unsigned int r;
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	if (t>254)
	{
		msgPop(msg,simple_pop,&t,sizeof(char),NULL);
		if ( t>254 )
		{
		/* eight octest */
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r += 16776960;
		}
		else
		{
		/* four octest */
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r*=256;
			msgPop(msg,simple_pop,&t,sizeof(char),NULL);
			r+=t;
			r+=65280;
		}
	
	}
	else
	{
	/* two octest */
		r = t;
		msgPop(msg,simple_pop,&t,sizeof(char),NULL);
		r*=256;
		r+=t;
	}	
	return r;
}


static void get_attribute(msg,a)
Msg * msg;
Attribute * a;
{
		msgPop(msg,simple_pop,&a->type,sizeof(char),NULL);
		if ( a->type != 0 )
		{
		
			msgPop(msg,simple_pop,&a->length,sizeof(char),NULL);
			if ( a->length != 0 )
			{
				msgPop(msg,simple_pop,a->value,a->length,NULL);
			}
		}
		else
		 	a->length = 0;
}

static void  get_attribute_list(self, msg,al)
XObj self;
Msg * msg;
Attribute_list * al;
{
	int x = 0;
	SState * ss = ((SState*)self->state);

	while(msgLen(msg) > 0)
	{
		if ( x >= MAX_ATTRIBUTE_LIST )
		{
			ss->state = DELETED_STATE;
			generate_error_message(self,2);
			xError("Size of Attribute list exeeds internal limit");
			return;
		}
		get_attribute(msg,&al->list[x]);
		x++;
	}
	al->length = x;
}

static void get_cookie_response(self, msg , str)
	XObj self;
	Msg * msg;
	Cookie_Response *str;
{ 
	char t;
	int  x = 0;
	SState * ss = ((SState*)self->state);
	xTrace0(photurisp, TR_EVENTS, "get_cookie_reponse entered");
	str->type = 1;
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	while(msgLen(msg) > 0)
	{
		if ( x >= MAX_GROUP_LIST )
		{
			ss->state = DELETED_STATE;
			generate_error_message(self,2);
			xError("Size of Goup list exeeds internal limit");
		}
		msgPop(msg,long_pop,&str->Offered_Schemes.list[x].index,2,NULL);
		x++;
	}
	str->Offered_Schemes.length=x; 	
	xTrace0(photurisp, TR_EVENTS, "get_cookie_response finished");
}

static void get_key_request(self, msg , str)
	XObj self;
	Msg * msg;
	Key_Request *str;
{
	char t;
 	xTrace0(photurisp, TR_EVENTS, "get_key_request entered");
	str->type = 2;
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	msgPop(msg,long_pop,&str->Scheme_Choice.index,2,NULL);
	/* lookup length of exchange value here */
	str->Initiator_Exchange_Value.length = get_size_of_variable_precision_number(msg);
	msgPop(msg,simple_pop,str->Initiator_Exchange_Value.value,
		roundup8(str->Initiator_Exchange_Value.length),NULL);

	get_attribute_list(self,msg,&str->Initiator_Offered_Attributes);
	
	xTrace0(photurisp, TR_EVENTS, "get_key_request_finished");
}

static void get_key_response(self, msg , str)
	XObj self;
	Msg * msg;
	Key_Response *str;
{
	char t;
 	xTrace0(photurisp, TR_EVENTS, "get_key_response entered");
	str->type = 3;
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	str->Responder_Exchange_Value.length = get_size_of_variable_precision_number(msg);
	msgPop(msg,simple_pop,str->Responder_Exchange_Value.value,
		roundup8(str->Responder_Exchange_Value.length),NULL);

 	get_attribute_list(self,msg,&str->Responder_Offered_Attributes);

 	xTrace0(photurisp, TR_EVENTS, "get_key_response finished");
}

static void get_signature_message(self, msg , str)
	XObj self;
	Msg * msg;
	Signature_Message *str;
{
	xTrace0(photurisp, TR_EVENTS, "get_signature_message entered");
	str->type = 4;
	msgPop(msg,long_pop,&str->Life_Time,3,NULL);
	msgPop(msg,long_pop,&str->Security_Parameter_Index,4,NULL);

	get_attribute(msg,&str->Signature_Choice);

	/* set here signature and certificate size */
	str->Certificate.length = get_size_of_variable_precision_number(msg);
	msgPop(msg,simple_pop,str->Certificate.value,
		roundup8(str->Certificate.length),NULL);
	str->Signature.length = get_size_of_variable_precision_number(msg);
	msgPop(msg,simple_pop,str->Signature.value,
		roundup8(str->Signature.length),NULL);

	get_attribute_list(self,msg,&str->Attribute_Choices);
	
 	xTrace0(photurisp, TR_EVENTS, "get_signature_message finished");
}

static void get_key_change(self, msg , str)
	XObj self;
	Msg * msg;
	Key_Change *str;
{
 	xTrace0(photurisp, TR_EVENTS, "get_key_change entered");
 	xTrace0(photurisp, TR_EVENTS, "get_key_change finished");
}

static void put_size_of_variable_precision_number(msg, sz)
	Msg * msg;
	unsigned int  sz;
{
	unsigned char t;
	if (sz > 16776959)
	{
		sz-=16776960;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		t=255;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		
	}
	else if ( sz > 65279 )
	{
		sz-=65280;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz=255;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
	}
	else
	{
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
		sz/=256;
		t=sz;
		msgPush(msg,simple_push,&t,sizeof(char),NULL);
	}
}


static void put_attribute(msg,a)
Msg * msg;
Attribute * a;
{
	unsigned char t = 0;
	if (a->length != 0)
		msgPush(msg,simple_push,a->value,a->length,NULL);
	t = a->length;
	if (a->type != 0 )
 		msgPush(msg,simple_push,&t,sizeof(char),NULL);
	t = a->type;
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
}

static void  put_attribute_list(msg,al)
Msg * msg;
Attribute_list * al;
{
	int x;
	for(x = al->length-1; x >= 0; x--)
	{
		put_attribute(msg,&al->list[x]);
	}
}

static void put_identity_verification(msg, shared_secret,offered_schemes,
		SPI_owner_exchange_val,
		SPI_user_exchange_val, SPI_owner_offered_attributes, 
		SPI_user_offered_attributes, type, life_time, SPI,
		Identification,  Attribute_Choice, Secret)
Msg * msg;
Public_value * shared_secret;
Groupe_list * offered_schemes;
Public_value * SPI_owner_exchange_val;
Public_value * SPI_user_exchange_val;
Attribute_list * SPI_owner_offered_attributes;
Attribute_list * SPI_user_offered_attributes; 
unsigned char *	type;
int * life_time;
tSPI * SPI;
Certificate * Identification;
Attribute_list * Attribute_Choice;
Key *	Secret;
{
	int x;
	msgPush(msg,simple_push,shared_secret->value,
		roundup8(shared_secret->length),NULL);
	msgPush(msg,simple_push,Secret->value,
		roundup8(Secret->length),NULL);

	put_attribute_list(msg,Attribute_Choice);

	msgPush(msg,simple_push,Identification->value,
		roundup8(Identification->length),NULL);
	msgPush(msg,long_push,SPI,4,NULL);
	msgPush(msg,long_push,life_time,3,NULL);
	msgPush(msg,simple_push,type,sizeof(char),NULL);

	put_attribute_list(msg,SPI_user_offered_attributes);

	put_attribute_list(msg,SPI_owner_offered_attributes);

	msgPush(msg,simple_push,SPI_user_exchange_val->value,
		roundup8(SPI_user_exchange_val->length),NULL);
	msgPush(msg,simple_push,SPI_owner_exchange_val->value,
		roundup8(SPI_owner_exchange_val->length),NULL);
	for(x=offered_schemes->length-1; x >= 0; x--)
	{
 		msgPush(msg,long_push,&offered_schemes->list[x].index,2,NULL);
	}
	msgPush(msg,simple_push,shared_secret->value,
		roundup8(shared_secret->length),NULL);

}


static void put_cookie_request(msg , str)
	Msg * msg;
	Cookie_Request *str;
{
	char type = str->type;
 	xTrace0(photurisp, TR_EVENTS, "put_cookie_request entered");
	msgPush(msg,simple_push,&type,sizeof(char),NULL);
 	xTrace0(photurisp, TR_EVENTS, "put_cookie_request finished");
}

static void put_cookie_response(msg , str)
	Msg * msg;
	Cookie_Response *str;
{
	char t = str->type;
	int  x;
 	xTrace0(photurisp, TR_EVENTS, "put_cookie_response entered");
	for(x=str->Offered_Schemes.length-1; x >= 0; x--)
	{
 		msgPush(msg,long_push,&str->Offered_Schemes.list[x].index,2,NULL);
	}
	t = 0;
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
	t =  str->type;
	msgPush(msg,simple_push,&t,sizeof(char),NULL);
 	xTrace0(photurisp, TR_EVENTS, "put_cookie_response finished");
}

static void put_key_request(msg , str)
	Msg * msg;
	Key_Request *str;
{
	char t = 0;
 	xTrace0(photurisp, TR_EVENTS, "put_key_request entered");

	put_attribute_list(msg,&str->Initiator_Offered_Attributes);

 	msgPush(msg,simple_push,str->Initiator_Exchange_Value.value,
		roundup8(str->Initiator_Exchange_Value.length),NULL);
	put_size_of_variable_precision_number(msg,str->Initiator_Exchange_Value.length);
 	msgPush(msg,long_push,&str->Scheme_Choice.index,2,NULL);
	t = 0;
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
	t =  str->type;
	msgPush(msg,simple_push,&t,sizeof(char),NULL);
  	xTrace0(photurisp, TR_EVENTS, "put_key_request finished");
}

static void put_key_response(msg , str)
	Msg * msg;
	Key_Response *str;
{
	char t = 0;
 	xTrace0(photurisp, TR_EVENTS, "put_key_response entered");

	put_attribute_list(msg,&str->Responder_Offered_Attributes);

 	msgPush(msg,simple_push,str->Responder_Exchange_Value.value,
		roundup8(str->Responder_Exchange_Value.length),NULL);
	put_size_of_variable_precision_number(msg,str->Responder_Exchange_Value.length);
	t = 0;
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
 	msgPush(msg,simple_push,&t,sizeof(char),NULL);
	t =  str->type;	
	msgPush(msg,simple_push,&t,sizeof(char),NULL);
 	xTrace0(photurisp, TR_EVENTS, "put_key_response finished");
}

static void put_signature_message(msg , str)
	Msg * msg;
	Signature_Message *str;
{
	char t = 0;
 	xTrace0(photurisp, TR_EVENTS, "put_signature_message entered");
	
	put_attribute_list(msg,&str->Attribute_Choices);

 	msgPush(msg,simple_push,str->Signature.value,roundup8(str->Signature.length),NULL);
	put_size_of_variable_precision_number(msg,str->Signature.length);
 	msgPush(msg,simple_push,str->Certificate.value,roundup8(str->Certificate.length),NULL);
	put_size_of_variable_precision_number(msg,str->Certificate.length);
	put_attribute(msg, &str->Signature_Choice);
	msgPush(msg,long_push,&str->Security_Parameter_Index,4,NULL);
	msgPush(msg,long_push,&str->Life_Time,3,NULL);
	t =  str->type;	
	msgPush(msg,simple_push,&t,sizeof(char),NULL);


  	xTrace0(photurisp, TR_EVENTS, "put_signature_message finished");
}

static void put_key_change(msg , str)
	Msg * msg;
	Key_Change *str;
{
	char t = str->type;
 	xTrace0(photurisp, TR_EVENTS, "put_key_change entered");
	msgPush(msg,simple_push,&t,sizeof(char),NULL);
 	xTrace0(photurisp, TR_EVENTS, "put_key_change_finished");
}

static int key_request_equal(x,y)
	Key_Request * x;
	Key_Request * y;
{
	int ret = 0;
 	xTrace0(photurisp, TR_EVENTS, "key_request_equal entered");
	if ((x->Scheme_Choice.index == y->Scheme_Choice.index)
	   && ( x->Initiator_Exchange_Value.length ==
		y->Initiator_Exchange_Value.length)
	   && (memcmp(x->Initiator_Exchange_Value.value,
		      y->Initiator_Exchange_Value.value, 
		      y->Initiator_Exchange_Value.length) == 0 )
	   && ( x->Initiator_Offered_Attributes.length ==
		y->Initiator_Offered_Attributes.length ))
		{
			int z;
			Attribute * a;
			Attribute * b;
			ret = 1;
			for ( z=0; (ret == 1) && (z <  x->Initiator_Offered_Attributes.length); x ++ )
			{
				a = &x->Initiator_Offered_Attributes.list[z];
				b = &y->Initiator_Offered_Attributes.list[z];
				if (( a->type != b->type )
				   || (a->length != b->length)
				   || (memcmp(a->value,b->value,a->length) != 0))
					ret = 0;
			}
		}
 	xTrace1(photurisp, TR_EVENTS, "key_request_equal finished %u",ret);
	return ret;
}

static int signature_message_equal(x,y)
	Signature_Message * x;
	Signature_Message * y;
{
	int ret = 0;
 	xTrace0(photurisp, TR_EVENTS, "signature_message_equal entered");
	if ((x->type == y->type)
	   && (x->Life_Time == y->Life_Time)
	   && (x->Security_Parameter_Index == y->Security_Parameter_Index)
	   && (x->Signature_Choice.type == y->Signature_Choice.type)
	   && (x->Signature_Choice.length == y->Signature_Choice.length)
	   && (memcmp(x->Signature_Choice.value,y->Signature_Choice.value,
		y->Signature_Choice.length)==0)
	   && ( x->Signature.length == y->Signature.length)
	   && (memcmp(x->Signature.value,y->Signature.value, y->Signature.length) == 0 )
	   && ( x->Certificate.length == y->Certificate.length)
	   && (memcmp(x->Certificate.value,y->Certificate.value, y->Certificate.length) == 0 )
	   && ( x->Attribute_Choices.length == y->Attribute_Choices.length ))
		{
			int z;
			Attribute * a;
			Attribute * b;
			ret = 1;
			for ( z=0; (ret == 1) && (z <  x->Attribute_Choices.length); x ++ )
			{
				a = &x->Attribute_Choices.list[z];
				b = &y->Attribute_Choices.list[z];
				if (( a->type != b->type )
				   || (a->length != b->length)
				   || (memcmp(a->value,b->value,a->length) != 0))
					ret = 0;
			}
		}
 	xTrace1(photurisp, TR_EVENTS, "signature_message_equal finished %u",ret);
	return ret;
}

void 	calculate_session_key(self, k, a, shared_secret, owner_signature, 
				user_signature, initiator_cookie, responder_cookie, spi)
	XObj self;
	Key * k;
	Attribute * a;
	Public_value * shared_secret;
	Signature * owner_signature, * user_signature;
	Cookie * initiator_cookie, *responder_cookie;
	tSPI * spi;
{
	Msg msg;
	SState * ss = ((SState*)self->state);
	PState * ps = ((PState*)self->myprotl->state);
	Part p[3];
	int x;
 	xTrace0(photurisp, TR_EVENTS, "calculate_session_key entered");
	partInit(p,3);
	msgConstructEmpty(&msg);
	switch(a->type)
	{
	/* MD5 */
	case 16:
	case 17:
	case 18:
	case 05:
		{
		XObj md5;
		XObj s;
		for(x=0; (x < ps->supported_attributes.length)
			 && 5 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("md5 as Signature_Choice md5 is not configured but used");
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		md5 = ps->attribute_protocols[x];

		partPush(p[LOCAL_PART],(void *)&ss->myhost,sizeof(IPhost));

		if ((s = xOpen(self->myprotl,self->myprotl,md5,p))==ERR_XOBJ)
		{
			xError("md5 as signature choice couldn't open md5 session");
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		msgPush(&msg,simple_push,shared_secret->value,roundup8(shared_secret->length),
			NULL);
		msgPush(&msg,long_push, spi, sizeof(tSPI), NULL);
		msgPush(&msg,simple_push,user_signature->value,roundup8(user_signature->length),
			NULL); 
		msgPush(&msg,simple_push,owner_signature->value,roundup8(owner_signature->length),
			NULL); 
		msgPush(&msg,simple_push,responder_cookie->value,COOKIE_SIZE,NULL);
		msgPush(&msg,simple_push,initiator_cookie->value,COOKIE_SIZE,NULL);
		msgPush(&msg,simple_push,shared_secret->value,roundup8(shared_secret->length),
			NULL);
		xPush(s,&msg);
		k->length=16*8;
		msgPop(&msg,simple_pop,k->value,16,NULL);
		xClose(s);
		break;
		}
	default:
		xError("Unsupported Hash_Choice in calculate_session_key");
		generate_error_message(self,0);
		ss->state = DELETED_STATE;
	}	
	xIfTrace(photurisp, TR_DETAILED)
	{
		int z;
		printf("photuris: Session Key: \n");
		if (ss->state != DELETED_STATE )
			for(z=0; z < roundup8(k->length) ; z ++)
				printf("%02x ",(unsigned char) k->value[z]);
		else
			printf("Session deleted\n");
		printf("\n--------\n");
	}
	msgDestroy(&msg);
 	xTrace0(photurisp, TR_EVENTS, "calculate_session_key finished");
}		

static void calculate_raw_key(self)
	XObj self;
{
	SState * ss = ((SState*)self->state);
 	xTrace0(photurisp, TR_EVENTS, "calculate_raw_key enterd");

	memcpy(ss->gxy.value,ss->gy.value,roundup8(ss->gy.length));
	ss->gxy.length=ss->gy.length;
	xControl(ss->gx_session,DHECKXLEAF_CALCULATE_SECRET,
			ss->gxy.value,roundup8(ss->gxy.length));
	xClose(ss->gx_session);
	ss->gx_session=NULL;

	xIfTrace(photurisp, TR_DETAILED)
	{
		int x;
		printf("photuris: Shared secret: \n");
		for(x=0; x < roundup8(ss->gxy.length) ; x ++)
			printf("%02x ",(unsigned char) ss->gxy.value[x]);
		printf("\n--------\n");
	}
	
 	xTrace0(photurisp, TR_EVENTS, "calculate_raw_key finished");
}



static void set_session_key(self)
	XObj self;
{
	KMB_MAPPING kmmb;
        XObj	    kms;
	Part	    kmp[1];
	Msg msg;
	SState * ss = ((SState*)self->state);
	PState * ps = ((PState*)self->myprotl->state);
	Key local,remote;

 	xTrace0(photurisp, TR_EVENTS, "set_session_key entered");

	calculate_session_key(self, &local, 
		(ss->state >= 129) ? &(ss->R_Attribute_Choices.list[0])
				: &(ss->I_Attribute_Choices.list[0]),
		&ss->gxy,
		&ss->outgoing_signature_message->Signature,
		&ss->incomming_signature_message->Signature,
		&ss->I_Cookie,
		&ss->R_Cookie,
		(ss->state >= 129 ) ? &ss->R_SPI : &ss->I_SPI);
	if ( ss->state == DELETED_STATE ) 
	{
		return;
	}

	calculate_session_key(self, &remote, 
		(ss->state < 129) ? &(ss->R_Attribute_Choices.list[0])
			: &(ss->I_Attribute_Choices.list[0]),
		&ss->gxy,
		&ss->incomming_signature_message->Signature,
		&ss->outgoing_signature_message->Signature,
		&ss->I_Cookie,
		&ss->R_Cookie,
		(ss->state < 129 ) ? &ss->R_SPI : &ss->I_SPI);
		
	if ( ss->state == DELETED_STATE )
	{ 
		return;
	}

	kmmb.local_host = ss->myhost;
	kmmb.local_algorithm = (ss->state >= 129) ? (ss->R_Attribute_Choices.list[0].type)
						: (ss->I_Attribute_Choices.list[0].type);
	kmmb.local_parameter = 0;
	kmmb.local_mode = (ss->mode<0)?((kmmb.remote_algorithm==5)?0:2):ss->mode;
	kmmb.local_timeout = (ss->state >= 129) ? (ss->R_lifetime) : (ss->I_lifetime);
	memset(kmmb.local_key,0,SAIDMGR_KEYSIZE);
	memcpy(&kmmb.local_key,&local.value, 
		(SAIDMGR_KEYSIZE < roundup8(local.length)? SAIDMGR_KEYSIZE : 
		roundup8(local.length)));
	kmmb.remote_host = ss->peerhost;
	kmmb.remote_said = (ss->state < 129) ? (ss->R_SPI) : (ss->I_SPI);
	kmmb.remote_algorithm = (ss->state < 129) ? (ss->R_Attribute_Choices.list[0].type)
						: (ss->I_Attribute_Choices.list[0].type);
	kmmb.remote_parameter = 0;
	kmmb.remote_mode = (kmmb.remote_algorithm==5)?0:2;
	kmmb.remote_timeout = (ss->state < 129) ? (ss->R_lifetime) : (ss->I_lifetime);
	memset(&kmmb.remote_key,0,SAIDMGR_KEYSIZE);
	memcpy(&kmmb.remote_key,&remote.value, 
		(SAIDMGR_KEYSIZE < roundup8(remote.length)? SAIDMGR_KEYSIZE : 
		roundup8(remote.length)));
	partInit(kmp,1);
     	partPush(kmp[0],((ss->state > 129) ? &(ss->R_SPI) : &(ss->I_SPI)),sizeof(XSaid));	
     	kms = xOpen(self->myprotl, self->myprotl, ps->km , kmp);
        if (kms == ERR_XOBJ) {ss->state = DELETED_STATE; return;}
     	xControl(kms, KM_SET, (char *) &kmmb, sizeof(KMB_MAPPING));
     	xClose(kms); 
	if( ss->state >= 129)
	{
		msgConstructEmpty(&msg);
		msgPush(&msg,long_push,(char *) (ss->state > 129) ? &(ss->R_SPI) : &(ss->I_SPI),
			sizeof(XSaid),NULL);
		xDemux(self,&msg);
		msgDestroy(&msg);
	}
	else
	{
		semSignal(&ss->initial_exchange);
	}

 	xTrace0(photurisp, TR_EVENTS, "set_session_key finished");

}

static	void calculate_signature(self,sig,cert,type)
	XObj self;
	Signature * sig;
	Certificate * cert;
	unsigned char type;
{
	XObj s;
	Msg    msg;
	SState * ss = ((SState*)self->state);
	PState * ps = ((PState*)self->myprotl->state);
	Part p[3];
	int x;
 	xTrace0(photurisp, TR_EVENTS, "calculate_signature entered");
	msgConstructEmpty(&msg);
	partInit(p,3);

	switch((ss->state>128)?ss->R_Signature_Choice.type:ss->I_Signature_Choice.type)
	{

	/* MD5 */
	case 05:
		{
		XObj md5;
		Key 	secret;
		char ibuf[MAX_CERTIFICATE_SIZE];
		int maxilength;
		memset(ibuf,0,MAX_CERTIFICATE_SIZE);
		secret.length=0;
		
		for(x=0; (x < ps->supported_attributes.length)
			 && 5 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			xError("md5 as Signature_Choice md5 is not configured but used");
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		md5 = ps->attribute_protocols[x];

	
	       	xTrace1(photurisp, TR_DETAILED, "calculate signature identity = %s", ss->user);

		xControl(ps->kmid,KM_HOSTSIZE,(char *)&maxilength,sizeof(int));
		memcpy(ibuf,ss->user,strlen(ss->user));
		partPush(p[LOCAL_PART],(void *)ibuf,maxilength);
		if ((s = xOpen(self->myprotl,self->myprotl,ps->kmid,p))!=ERR_XOBJ)
		{
			xControl(s,KM_KEYSIZE,(char *)&maxilength,sizeof(int));
			if ((maxilength =xControl(s,KM_RESOLVE,secret.value,maxilength))!=XK_FAILURE)
			{
				xTrace1(photurisp, TR_DETAILED, 
					"calculate_signature key size %u",maxilength );
				/* put in shared secret */
				secret.length=maxilength*8;
			} 
			else
			{
				generate_error_message(self,0);
				xError("No valid identity");
				ss->state=DELETED_STATE;
			}
			xClose(s);
		}
		else
		{
			generate_error_message(self,0);
			xError("No valid identity");
			ss->state=DELETED_STATE;
		}
		
		cert->length=strlen(ss->user)*8;
		memcpy(cert->value,ss->user,roundup8(cert->length));
		
		put_identity_verification(&msg, &ss->gxy, &ss->Offered_Schemes,
			&ss->gx, &ss->gy, 
			(ss->state > 128)? &ss->R_Offered_Attributes:&ss->I_Offered_Attributes,
			(ss->state > 128)? &ss->I_Offered_Attributes:&ss->R_Offered_Attributes,
			&type, 
			(ss->state > 128)? &ss->R_lifetime:&ss->I_lifetime,
			(ss->state > 128)? &ss->R_SPI:&ss->I_SPI, cert,
 			(ss->state > 128)? &ss->R_Attribute_Choices:&ss->I_Attribute_Choices,
 			&secret);
		if ((s = xOpen(self->myprotl,self->myprotl,md5,p))==ERR_XOBJ)
		{
			xError("md5 as signature choice couldn't open md5 session");
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			break;
		}
		xPush(s,&msg);
		sig->length=16*8;
		msgPop(&msg,simple_pop,sig->value,16,NULL);
		xClose(s);
		break;
		}
	default:
		xError("Unsupported Signature_Choice in calculate_signature");
		generate_error_message(self,0);
		ss->state = DELETED_STATE;
	}	
	msgDestroy(&msg);
 	xTrace0(photurisp, TR_EVENTS, "calculate_signature finished");
}

static	int check_signature(self,sig,cert,type)
	XObj self;
	Signature * sig;
	Certificate * cert;
	unsigned char type;
{
	XObj s;
	Msg    msg;
	SState * ss = ((SState*)self->state);
	PState * ps = ((PState*)self->myprotl->state);
	Part p[3];
	int x;
 	xTrace0(photurisp, TR_EVENTS, "check_signature entered");
	msgConstructEmpty(&msg);
	partInit(p,3);


	switch((ss->state<=128)?ss->R_Signature_Choice.type:ss->I_Signature_Choice.type)
	{

	/* MD5 */
	case 05:
		{
		XObj md5;
		Key 	secret;
		Signature sig2;
		char ibuf[MAX_CERTIFICATE_SIZE];
		int maxilength;
		memset(ibuf,0,MAX_CERTIFICATE_SIZE);
		secret.length=0;

		if (sig->length != 16*8) /* can't be an MD5 hash */
			return 0;

		for(x=0; (x < ps->supported_attributes.length)
			 && 5 != ps->supported_attributes.list[x].type; x++);
		if ( x >=  ps->supported_attributes.length )
		{
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			xError("md5 as Signature_Choice md5 is not configured but used");
			return -1;
			break;
		}
		md5 = ps->attribute_protocols[x];

	       	xTrace1(photurisp, TR_DETAILED, "check signature identity = %s", cert->value);
	
		xControl(ps->kmid,KM_HOSTSIZE,(char *)&maxilength,sizeof(int));
		memcpy(ibuf,cert->value,roundup8(cert->length));
		partPush(p[LOCAL_PART],(void *)ibuf,maxilength);
		if ((s = xOpen(self->myprotl,self->myprotl,ps->kmid,p))!=ERR_XOBJ)
		{
			xControl(s,KM_KEYSIZE,(char *)&maxilength,sizeof(int));
			
			if ((maxilength = xControl(s,KM_RESOLVE,secret.value,maxilength))!=XK_FAILURE)
			{
				xTrace1(photurisp, TR_DETAILED, 
					"check_signature key size %u",maxilength );
				/* put in shared secret */
				secret.length=maxilength*8;
			}
			else
			{
				generate_error_message(self,3);
				ss->state = DELETED_STATE;
				xTrace0(photurisp, TR_EVENTS, "check_signature no valid certificat");
				return -1;
			}
			xClose(s);
		}
		else
		{
			generate_error_message(self,3);
			ss->state = DELETED_STATE;
			xTrace0(photurisp, TR_EVENTS, "check_signature identity unknown");
			return -1;
		}

				
		put_identity_verification(&msg, &ss->gxy, &ss->Offered_Schemes,
			&ss->gy, &ss->gx, 
			(ss->state <= 128)? &ss->R_Offered_Attributes:&ss->I_Offered_Attributes,
			(ss->state <= 128)? &ss->I_Offered_Attributes:&ss->R_Offered_Attributes,
			&type, 
			(ss->state <= 128)? &ss->R_lifetime:&ss->I_lifetime,
			(ss->state <= 128)? &ss->R_SPI:&ss->I_SPI, cert,
 			(ss->state <= 128)? &ss->R_Attribute_Choices:&ss->I_Attribute_Choices,
 			&secret);
			
		if ((s = xOpen(self->myprotl,self->myprotl,md5,p))==ERR_XOBJ)
		{
			generate_error_message(self,0);
			ss->state = DELETED_STATE;
			xError("md5 as signature choice couldn't open md5 session");
			break;
		}
		xPush(s,&msg);
		sig2.length=16*8;
		msgPop(&msg,simple_pop,sig2.value,16,NULL);
		xClose(s);
        	xTrace1(photurisp, TR_EVENTS, "check_signature result %u",(0==memcmp(sig->value,sig2.value,16)));
		return (0==memcmp(sig->value,sig2.value,16));
		break;
		}
	default:
		xError("Unsupported Signature_Choice in check_signature");
		ss->state = DELETED_STATE;
		return 0;
	}	
	msgDestroy(&msg); 	
        xTrace0(photurisp, TR_EVENTS, "check_signature finished");
	return 1;
}

static void do_cookie_request(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Cookie_Request rstr;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_cookie_request entered");
	msgConstructEmpty(&rmsg);
	rstr.type=0;
	put_cookie_request(&rmsg,&rstr);
	xPush(xGetDown(self,0),&rmsg);
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_cookie_request finished");
}

static void do_cookie_response(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Cookie_Response rstr;
 	PState * ps = (PState *) self->myprotl->state;
 
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_cookie_response entered");
	msgConstructEmpty(&rmsg);
	rstr.type=1;
	rstr.Offered_Schemes = ps->supported_groups;
	put_cookie_response(&rmsg,&rstr);
	xPush(xGetDown(self,0),&rmsg);
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_cookie_response finished");

}

static void do_key_request(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Cookie_Response istr;
	Key_Request    * rstr;
	SState * ss = (SState *) self->state;
 	PState * ps = (PState *) self->myprotl->state;
	int x;
        int y;

 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_request entered");
	msgConstructEmpty(&rmsg);
	if (ss->incomming_key_message != NULL)
	    {
		put_key_request(&rmsg,ss->incomming_key_message);
		xPush(xGetDown(self,0),&rmsg);
		return;
	    }

	get_cookie_response(self,msg,&istr);
	if (ss->state == DELETED_STATE)
	{
		return;
	}
	xControl(xGetDown(self,0),COOKIE_PHOTURIS_INITIATOR_COOKIE,ss->I_Cookie.value, COOKIE_SIZE);
	xControl(xGetDown(self,0),COOKIE_PHOTURIS_RESPONDER_COOKIE,ss->R_Cookie.value, COOKIE_SIZE);

	rstr = (Key_Request *) xMalloc(sizeof(Key_Request));

	/*choose schema */
	y = 0;
	x = 0;
	while ( (y < istr.Offered_Schemes.length )
	      && (istr.Offered_Schemes.list[y].index != ps->supported_groups.list[x].index))
	  {
	    x=0;
	    while((x< ps->supported_groups.length) 
	      && (istr.Offered_Schemes.list[y].index != ps->supported_groups.list[x].index))
		    x++;
 	   if (istr.Offered_Schemes.list[y].index != ps->supported_groups.list[x].index) y++;
         }
	if (istr.Offered_Schemes.list[y].index == ps->supported_groups.list[x].index)
	    {
		ss->Group_index = ps->supported_groups.list[x];
		/* for debugging only will only work with debugging 
		    graph.comp   ss->Group_index.index = 2; x = 1; */
		ss->Offered_Schemes = istr.Offered_Schemes;
		xControl(ps->group_protocols[x],DHECKXLEAF_VALUE_SIZE,(char *)&ss->gx.length,sizeof(int));
		xControl(ps->group_sessions[x],DHECKXLEAF_PUBLIC_VALUE,
			ss->gx.value,roundup8(ss->gx.length));
		xDuplicate(ps->group_sessions[x]);
		ss->gx_session = ps->group_sessions[x];
		xDuplicate(ss->gx_session);
		ss->I_Offered_Attributes = ps->supported_attributes;

		rstr->type=2;
		rstr->Scheme_Choice = ss->Group_index;		
		rstr->Initiator_Exchange_Value = ss->gx;
		rstr->Initiator_Offered_Attributes = ss->I_Offered_Attributes;
		put_key_request(&rmsg,rstr);
		xPush(xGetDown(self,0),&rmsg);

		ss->incomming_key_message = rstr;
		
	    }
	else
	    {
		generate_error_message(self,0);
		ss->state = DELETED_STATE;
	    }
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_request finished");
	
	 
}

static void do_key_response(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Key_Request * istr;
	Key_Response * rstr;
	SState * ss = (SState *) self->state;
 	PState * ps = (PState *) self->myprotl->state;
	int x;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_response entered");
	istr = (Key_Request * ) xMalloc(sizeof(Key_Request));
	rstr = (Key_Response * ) xMalloc(sizeof(Key_Response));
	get_key_request(self,msg,istr);
	msgConstructEmpty(&rmsg);
	

	/* if I get the same request again send the same response */
	if ((ss->incomming_key_message != NULL )
	   && (key_request_equal(istr, ss->incomming_key_message)))
	{
		put_key_response(&rmsg, ss->outgoing_key_message);
		xPush(xGetDown(self,0),&rmsg);
		xFree((char *) istr);
		xFree((char *) rstr);
		return;
	}	
	if (ss->state == DELETED_STATE)
	{
		return;
	}
	xControl(xGetDown(self,0),COOKIE_PHOTURIS_INITIATOR_COOKIE,ss->I_Cookie.value, COOKIE_SIZE);
	xControl(xGetDown(self,0),COOKIE_PHOTURIS_RESPONDER_COOKIE,ss->R_Cookie.value, COOKIE_SIZE);
        x=0;
	while((x< ps->supported_groups.length) 
	      && (istr->Scheme_Choice.index != ps->supported_groups.list[x].index))
		    x++;
	if (x< ps->supported_groups.length)
	    {
		ss->Offered_Schemes = ps->supported_groups; /* bug shouldn't be assignet here */
		ss->Group_index = istr->Scheme_Choice;
		xControl(ps->group_protocols[x],DHECKXLEAF_VALUE_SIZE,(char *)&ss->gx.length,sizeof(int));
		xControl(ps->group_sessions[x],DHECKXLEAF_PUBLIC_VALUE,
			ss->gx.value,roundup8(ss->gx.length));
		xDuplicate(ps->group_sessions[x]);
		ss->gx_session = ps->group_sessions[x];
		xDuplicate(ss->gx_session);
		ss->gy = istr->Initiator_Exchange_Value;
		ss->I_Offered_Attributes = istr->Initiator_Offered_Attributes;
		ss->R_Offered_Attributes = ps->supported_attributes;

		rstr->type=3;
		rstr->Responder_Exchange_Value = ss->gx;
		rstr->Responder_Offered_Attributes = ss->R_Offered_Attributes;
		put_key_response(&rmsg,rstr);
		xPush(xGetDown(self,0),&rmsg);

  		ss->incomming_key_message = istr;
		ss->outgoing_key_message = rstr;		

		/* calculate key here based on chosen group*/
		calculate_raw_key(self);		
				
	    }
	else
	    {
		generate_error_message(self,0);
		ss->state = DELETED_STATE;
	    }
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_response finished");
	
	
}

static void do_signature_message_initiator(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Key_Response istr;
	Signature_Message * rstr;
	SState * ss = (SState *) self->state;
 	Signature	s;
	Certificate	c;

 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_signature_message_initiator entered");

	msgConstructEmpty(&rmsg);
	if (ss->outgoing_signature_message != NULL)
	    {
		put_signature_message(&rmsg,ss->outgoing_signature_message);
		encrypt_signature_message(&rmsg,self);
		xPush(xGetDown(self,0),&rmsg);
		return;
	    }

	get_key_response(self,msg,&istr);
	if (ss->state == DELETED_STATE) return;

	rstr = (Signature_Message *) xMalloc(sizeof(Signature_Message));
	
	/* missing check if anonymity choice is supported */
	ss->gy = istr.Responder_Exchange_Value;
	ss->R_Offered_Attributes = istr.Responder_Offered_Attributes;

	/* calculate key here based on chosen group*/
	calculate_raw_key(self);		

	/* chose I_Signature_Choice  */
	/* I_Attribute_Choices I_SPI I_lifetime should allreay be set */

	ss->I_Signature_Choice.type = 5; /* always MD5 */
	ss->I_Signature_Choice.length = 0; 


	/* calculate signature */
	
	calculate_signature(self,&s,&c,4);

	rstr->type=4;
	rstr->Life_Time = ss->I_lifetime;
	rstr->Security_Parameter_Index = ss->I_SPI;
	rstr->Signature_Choice = ss->I_Signature_Choice;
	rstr->Signature = s;
	rstr->Certificate = c;
	rstr->Attribute_Choices = ss->I_Attribute_Choices;
	put_signature_message(&rmsg,rstr);
	encrypt_signature_message(&rmsg,self);
	if (ss->state == DELETED_STATE) return;
	xPush(xGetDown(self,0),&rmsg);
	
	ss->outgoing_signature_message = rstr;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_signature_message_initiator finished");
	
}

static void do_signature_message_responder(self,msg)
	XObj self;
	Msg * msg;
{
	Msg rmsg;
	Signature_Message * istr;
	Signature_Message * rstr;
	SState * ss = (SState *) self->state;
	PState * ps = (PState *) self->myprotl->state;
 	Signature	s;
	Certificate	c;
	int		x;
	XAlgorithm	*y;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_signature_message_responder eneterd");
	istr = (Signature_Message *) xMalloc(sizeof(Signature_Message));
	rstr = (Signature_Message *) xMalloc(sizeof(Signature_Message));
	decrypt_signature_message(msg,self);
	if (ss->state == DELETED_STATE) return;
	get_signature_message(self,msg,istr);
	if (ss->state == DELETED_STATE) return;
	msgConstructEmpty(&rmsg);
	
	/* if I get the same request again send the same response */
	if ((ss->incomming_signature_message != NULL )
	   && (signature_message_equal(istr, ss->incomming_signature_message)))
	{
		put_signature_message(&rmsg,ss->outgoing_signature_message);
		encrypt_signature_message(&rmsg,self);
		xPush(xGetDown(self,0),&rmsg);
		xFree((char *) istr);
		xFree((char *) rstr);
		return;
	}

	ss->I_lifetime = istr->Life_Time;
	ss->I_SPI = istr->Security_Parameter_Index;
	ss->I_Signature_Choice = istr->Signature_Choice;
	s = istr->Signature;
	c = istr->Certificate;
	ss->I_Attribute_Choices = istr->Attribute_Choices;


	/* chose  R_Signature_Choice R_Attribute_Choices */
	/* R_SPI R_lifetime should allreay be set */

	ss->R_Signature_Choice.type = 5; /* always MD5 */
	ss->R_Signature_Choice.length = 0; 

	for( x = 0; x < ss->I_Attribute_Choices.length; x++ )
	{
		XAlgorithm a = ss->I_Attribute_Choices.list[x].type;
		if ( mapResolve(ps->algoMap,&a,&y)
			 == XK_FAILURE)
		{
			xError(
			"photuris:: do_signature_message_responder:: attribute_choices not allowed");
			ss->state = DELETED_STATE;
			return;
		}
	}
	ss->R_Attribute_Choices = ss->I_Attribute_Choices;

	if (check_signature(self,&s,&c,4))
	{
		/* calculate signature */
	
		calculate_signature(self,&s,&c,4);

		rstr->type=4;
		rstr->Life_Time = ss->R_lifetime;
		rstr->Security_Parameter_Index = ss->R_SPI;
		rstr->Signature_Choice = ss->R_Signature_Choice;
		rstr->Signature = s;
		rstr->Certificate = c;
		rstr->Attribute_Choices = ss->R_Attribute_Choices;
		put_signature_message(&rmsg,rstr);
		encrypt_signature_message(&rmsg,self);
		if (ss->state == DELETED_STATE) return;
		xPush(xGetDown(self,0),&rmsg);
		
		ss->incomming_signature_message = istr;
		ss->outgoing_signature_message = rstr;
		/*Put session key  into the SAIDMGR */
		set_session_key(self);
	}
	else
	    {
		generate_error_message(self,3);
		ss->state = DELETED_STATE;
	    }
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_signature_message_responder finished");
}
			
static void do_received_signature_message_responseer(self,msg)
	XObj self;
	Msg * msg;
{
	Signature_Message *  istr;
	SState * ss = (SState *) self->state;
 	Signature	s;
	Certificate	c;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_received_signature_message_responseer entered");
	decrypt_signature_message(msg,self);
	if (ss->state == DELETED_STATE) return;
	istr = (Signature_Message *) xMalloc(sizeof(Signature_Message));
	get_signature_message(self, msg,istr);
	if (ss->state == DELETED_STATE) return;
	ss->R_lifetime = istr->Life_Time;
	ss->R_SPI = istr->Security_Parameter_Index;
	ss->R_Signature_Choice = istr->Signature_Choice;
	s = istr->Signature;
	c = istr->Certificate;
	ss->R_Attribute_Choices = istr->Attribute_Choices;
	ss->incomming_signature_message = istr;

	if (check_signature(self,&s,&c,4))
	{
		
		/*Put session key  into the SAIDMGR */
		set_session_key(self);
	}
	else
	    {
		generate_error_message(self,3);
		ss->state = DELETED_STATE;
	    }
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_received_signature_message_responseer finished");
}

static void do_key_change(self,msg)
	XObj self;
	Msg * msg;
{
 	SState * ss = (SState *) self->state;
	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_change entered");
	ss->state = DELETED_STATE;
	generate_error_message(self,2);
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_key_change finished");
}

static void do_error_message(self,msg)
	XObj self;
	Msg * msg;
{
	unsigned char t;
	SState * ss = (SState *) self->state;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_error_message entered");
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	msgPop(msg,simple_pop,&t,sizeof(char),NULL);
	ss->state = DELETED_STATE;
	xTrace1(photurisp, TR_MAJOR_EVENTS, "do_error_message error_reason %u",t);	
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "do_error_message finished");
}

static void	generate_error_message(self,reason)
	XObj self;
	unsigned char reason;
{
	Msg msg;
	unsigned char type = 7;
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "generate_error_message entered");
	msgConstructEmpty(&msg);
	msgPush(&msg,simple_push,&reason,sizeof(char),NULL);
	msgPush(&msg,simple_push,&type,sizeof(char),NULL);
	xPush(xGetDown(self,0),&msg);
	xTrace1(photurisp, TR_MAJOR_EVENTS, "generate_error_message %u",reason);	
 	xTrace0(photurisp, TR_MAJOR_EVENTS, "generate_error_message finished");
}

static xkern_return_t photurisPop(self, lls, msg, hdr) 	
	XObj  self,lls;
	Msg * msg;
	void * hdr;
{
	int event;
	SState * ss = (SState *) self->state;
	PState * ps = (PState *) self->myprotl->state;
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisPop entered");
	if (!xIsSession(self)) 
	{ xError("XObject not a session state in next_state()");
	  return XK_FAILURE;
        }
	semWait(&ss->process);
	event = *((int *)hdr);
 	xTrace2(photurisp, TR_DETAILED, "photurisPop state %u event %u",ss->state,event);
	switch (ss->state)
	{
	case DELETED_STATE:
		switch (event)
		{
		case COOKIE_REQUEST:
		case COOKIE_RESPOND:
		case KEY_REQUEST:
		case KEY_RESPOND:
		case SIGNATURE_MESSAGE:
			generate_error_message(self,1);
			break;
		default:
			xTrace0(photurisp,TR_ERRORS,"Unexpected event during DELETED_STATE");
		}
		break;
	case IDLE_STATE:
		switch (event)
		{
		case XOPEN:
			ss->state = INITIATOR_COOKIE_REQUEST;
			do_cookie_request(self,msg);
			ss->timeout=1;
			break;
		case COOKIE_REQUEST:
			do_cookie_response(self,msg);
			xClose(self);
			break;
		case KEY_REQUEST:
			ss->state = RESPONDER_KEY_RESPOND;
			do_key_response(self,msg);
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case TIMEOUT:
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			xTrace0(photurisp,TR_ERRORS,"Unexpected event during IDLE_STATE");
		}
		break;
	case INITIATOR_COOKIE_REQUEST:
		switch (event)
		{
		case COOKIE_RESPOND:
			ss->state = INITIATOR_KEY_REQUEST;
			do_key_request(self,msg);
			ss->timeout=1;
			break;
		case TIMEOUT:
			do_cookie_request(self,msg);
			ss->timeout=1;
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Initiator ignors other events */
		}
		break;
	case INITIATOR_KEY_REQUEST:
		switch (event)
		{
		case KEY_RESPOND:
			ss->state = INITIATOR_SIGNATURE_MESSAGE;
			do_signature_message_initiator(self,msg);
			ss->timeout=1;
			break;
		case TIMEOUT:
			do_key_request(self,msg);
			ss->timeout=1;
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Initiator ignors other events */
		}
		break;
	case INITIATOR_SIGNATURE_MESSAGE:
		switch (event)
		{
		case SIGNATURE_MESSAGE:
			ss->state = INITIATOR_SIGNATURE_MESSAGE_FINISHED;
			do_received_signature_message_responseer(self,msg);
			ss->timeout=-1;
			ss->global_timeout=ps->global_timeout;
			break;
		case TIMEOUT:
			do_signature_message_initiator(self,msg);
			ss->timeout=1;
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Initiator ignors other events */
		}
		break;
	case INITIATOR_SIGNATURE_MESSAGE_FINISHED:
		switch (event)
		{
		case KEY_CHANGE:
			do_key_change(self,msg);
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Initiator ignors other events */
		}
		break;
	case RESPONDER_KEY_RESPOND:
		switch (event)
		{
		case KEY_REQUEST:
			ss->state = RESPONDER_KEY_RESPOND;
			do_key_response(self,msg);
			break;			
		case SIGNATURE_MESSAGE:
			ss->state = RESPONDER_SIGNATURE_MESSAGE;
			do_signature_message_responder(self,msg);
			ss->timeout=-1;
			ss->global_timeout=ps->global_timeout;
			break;			
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Responder ignors other events */
		}
		break;
	case RESPONDER_SIGNATURE_MESSAGE:
		switch (event)
		{
		case SIGNATURE_MESSAGE:
			do_signature_message_responder(self,msg);
			break;			
		case KEY_CHANGE:
			do_key_change(self,msg);
			break;
		case ERROR_MESSAGE:
			do_error_message(self,msg);
			break;
		case GLOBAL_TIMEOUT:
			ss->state = DELETED_STATE;
			break;
		default:
			break;	/* Responder ignors other events */
		}
		break;
	default:
		xError("Illegal session state in next_state()");
	} 
	if (ss->state == DELETED_STATE)
	{
		ss->timeout=-1;
		ss->global_timeout=-1;
		semSignal(&ss->initial_exchange);
	}
	semSignal(&ss->process);
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisPop finished");
	return XK_SUCCESS;
}


static XObj photurisOpen(self, hlp, hlpType, part)
     XObj self, hlp, hlpType;
     Part *part;
{
 	PState	*ps = (PState *)self->state;
  	int event;
	XSaid  lsaid;
	XIpsecMode mode;
	XAlgorithm algo,*x;
	char * user;
	XObj s=0;
	XObj lls;
	long  serverPort = 468;
	long  localPort;
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisOpen entered");
	xControl(xGetDown(self,0),UDP_GETFREEPORTNUM,(char*)&localPort,sizeof(long));
	partPush(part[REMOTE_PART],&serverPort,sizeof(long));
	partPush(part[LOCAL_PART],&localPort,sizeof(long));
    	lsaid = *((XSaid *)partPop(part[SPI_PART]));
    	algo = *((XAlgorithm *)partPop(part[SPI_PART]));
    	mode = *((XIpsecMode *)partPop(part[SPI_PART]));
	user = (char *) partPop(part[SPI_PART]);
	if ( user == NULL)
	{
		xError("Photuris: No identity provided\n");
		return ERR_XOBJ;
	}
	if ( mapResolve(ps->algoMap,&algo,&x) == XK_FAILURE)
		return ERR_XOBJ;
	lls = xOpen(self,self,xGetDown(self,0),part);
	s = photurisCreateSessn(hlpType, self, lls);
	((SState *)s->state)->I_SPI = lsaid;
	((SState *)s->state)->I_lifetime = ps->default_key_lifetime;
	((SState *)s->state)->I_Attribute_Choices.length = 1;
	((SState *)s->state)->I_Attribute_Choices.list[0].type = algo;
	((SState *)s->state)->I_Attribute_Choices.list[0].length = 0;
	((SState *)s->state)->mode = mode;
	((SState *)s->state)->local_port = localPort;
	memcpy(((SState *)s->state)->user,user,strlen(user));
	event = XOPEN;
	xPop(s, lls, NULL, &event);
	semWait(&(((SState *)s->state)->initial_exchange));
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisOpen finished");
	if  ((((SState *)s->state)->state) != DELETED_STATE)
		return s;
	else
	{
		xClose(s);
		return ERR_XOBJ;
	}
}



static XObj	 
photurisCreateSessn( hlp, self, lls )
    XObj	hlp, self, lls;
{
    	SState  * ss;
	XObj s;
  	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisCreateSessn entered");
   	s = xCreateSessn(photurisSessnInit, hlp, hlp, self, 1 , &lls);
	xDuplicate(s);
    	s->binding = (Bind) mapBind(((PState *)self->myprotl->state)->actMap,&lls , s);
	ss =(SState *) xMalloc(sizeof(SState));
    	bzero((void *)ss, sizeof(SState));
    	ss->state = IDLE_STATE;
 	semInit(&ss->process,1);
 	semInit(&ss->initial_exchange,0);
    	s->state = (void *)ss;
	ss->global_timeout = ((PState *)self->myprotl->state)->global_timeout;
	ss->timeout = -1;
	ss->gx_session=NULL;
	xControl(lls,GETMYHOST,(void *)&ss->myhost,sizeof(IPhost));
	xControl(lls,GETPEERHOST,(void *)&ss->peerhost,sizeof(IPhost));
	
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisCreateSessn finished");
	return s;
}


static xkern_return_t	photurisOpenDone( self, lls, llp, hlpType  )
     XObj self, lls, llp, hlpType;
{
	XObj s=0;
 	xTrace1(photurisp, TR_GROSS_EVENTS, "photurisOpenDone entered lls%08x",lls);
	if ( mapResolve(((PState *) self->myprotl->state)->actMap, &lls, &s) == XK_FAILURE ) 
  	{ 
		s=photurisCreateSessn(hlpType, self, lls ); 
		((SState *)s->state)->R_SPI = (int) xOpenDone(hlpType,s,self);
		((SState *)s->state)->R_lifetime = ((PState *)self->myprotl->state)->default_key_lifetime;
		((SState *)s->state)->mode = -1;
		xDuplicate(lls);
		memcpy(((SState *)s->state)->user, ((PState *)s->myprotl->state)->user, 
			strlen(((PState *)s->myprotl->state)->user));
 	}
	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisOpenDone finished");
	return XK_SUCCESS;
}

static xkern_return_t photurisOpenEnable( self, hlp, hlpType, pi)
XObj self,hlp, hlpType;
Part * pi;
{
	static long	serverPort = 468; /* according to the draft */
	Part p[3];
	XAlgorithm  algo;
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisOpenEnable entered");
	algo = *((XAlgorithm *) partPop(pi[LOCAL_PART]));
	mapBind(((PState *)self->myprotl->state)->algoMap,&algo , &algo);
    	partInit(p, 1);
    	partPush(*p, ANY_HOST, 0);
    	partPush(*p, &serverPort, sizeof(long));
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisOpenEnable finished");
	return xOpenEnable(self,hlpType,xGetDown(self,0),p);
}

static xkern_return_t photurisDemux(self, lls, msg)
     XObj self, lls;
     Msg  *msg;
{
   	char t;
	int event;
	XObj s=0;
  	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisDemux entered");

 	msgPop(msg, simple_pop, &t, sizeof(char),NULL);
	event = t;
  
	if ( mapResolve(((PState *) self->myprotl->state)->actMap, &lls, &s) == XK_FAILURE ) 
  	{ 
		xTrace0(photurisp,TR_GROSS_EVENTS,"photurisDemux unknown lls");
		return XK_FAILURE;
	}

 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisDemux finished");
	return xPop(s, lls, msg, &event);
}


static xkern_return_t romservername(XObj, char **, int, int, void *);
static xkern_return_t romgrouptimeout(XObj, char **, int, int, void *);
static xkern_return_t romexchangetimeout(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[] = {{ "servername", 3, romservername },
				{ "grouptimeout", 3, romgrouptimeout },
				{ "exchangetimeout", 3, romexchangetimeout}
				 };
static xkern_return_t
romservername(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  /* will set or erase keyfile name */
  if (fields > 2)
    memcpy(((PState *)(self->state))->user,ptr[2],strlen(ptr[2])+1);
  else
    Kabort("No server identity set");
  printf("photuris romoption setservname %s \n",ptr[2]);
  return XK_SUCCESS;
}

static xkern_return_t
romgrouptimeout(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  /* will set or erase keyfile name */
  if (fields > 2)
  {
    ((PState *)(self->state))->group_timeout_ini = atoi(ptr[2]);
    ((PState *)(self->state))->group_timeout = ((PState *)(self->state))->group_timeout_ini ;
     printf("photuris romoption grouptimeout %u \n",((PState *)(self->state))->group_timeout_ini);
  }
  else
    Kabort("Not enough arguments for grouptimeout");
  return XK_SUCCESS;
}

static xkern_return_t
romexchangetimeout(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  /* will set or erase keyfile name */
  if (fields > 2)
  {
    ((PState *)(self->state))->global_timeout = atoi(ptr[2]);
     printf("photuris romoption exchangetimeout %u \n",((PState *)(self->state))->global_timeout);
  }
  else
    Kabort("Not enough arguments for exchangetimeout");
  return XK_SUCCESS;
}

void photuris_init( self )
     XObj self;
{
  	PState *ps;
	int	j = 1;
	int	sl= 0;
	int	al= 0;
	int	random;

 	xTrace0(photurisp, TR_GROSS_EVENTS, "photuris_init entered");

   	ps = (PState *) xMalloc(sizeof(PState));
	ps->user[0] = 0;
     	ps->actMap = mapCreate(PHOTURIS_ACTIVE_MAP_SZ, sizeof(XObj));
     	ps->algoMap = mapCreate(PHOTURIS_ALGO_MAP_SZ, sizeof(XAlgorithm));
	ps->default_key_lifetime=3600;
	ps->timeout_time=10000000;
	ps->group_timeout_ini=60;
	ps->group_timeout = ps->group_timeout_ini;
	ps->global_timeout = 30;
	ps->km =  xGetDown(self, j++);
	ps->kmid =  xGetDown(self, j++);
	while ( (sl < self->numdown - 3 ) && (xGetDown(self, j)->instName != NULL) &&
		(xGetDown(self, j)->instName[0] == 'g'))
	{
		ps->group_protocols[sl] = xGetDown(self, j++);
		sl ++;
	}
	while ( (al < self->numdown - 3 - sl ) && (xGetDown(self, j)->instName != NULL) &&
		(xGetDown(self, j)->instName[0] == 'a'))
	{
		ps->attribute_protocols[al] = xGetDown(self, j++);
		al ++;
	}
	/* should be set from KM */
	ps->supported_groups.length = sl;
	for(j=0; j < sl; j++)
	{
		ps->group_sessions[j]=xOpen(self,self,ps->group_protocols[j],NULL);
		ps->supported_groups.list[j].index=atoi(&ps->group_protocols[j]->instName[1]);
		xTrace1(photurisp, TR_GROSS_EVENTS, "photuris_init found algorithm for group %u"
			,ps->supported_groups.list[j].index);
	}
	ps->supported_attributes.length = al;
	for(j=0; j < al; j++)
	{
		ps->supported_attributes.list[j].length = 0;	/* bug no attribute with length
									!= 0 yet*/ 
	
		ps->supported_attributes.list[j].type=
			atoi(&ps->attribute_protocols[j]->instName[1]);
		xTrace1(photurisp, TR_GROSS_EVENTS, "photuris_init found algorithm for attribute %u"
			,ps->supported_attributes.list[j].type);
	}
	
    	self->state = (VOID *)ps;

  	self->open = photurisOpen;
 	self->openenable = photurisOpenEnable;
  	self->opendone = photurisOpenDone;
  	self->pop = photurisPop;
  	self->demux = photurisDemux;

 	random = rand();
	xControl(xGetDown(self,0),COOKIE_PHOTURIS_SET_RANDOM_VALUE,(char *)&random,sizeof(int));
    	findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(rmopts[0]), 0);
	if (ps->user[0] == 0)
		Kabort("No server identity set");
	/* accept any connection */
	evDetach(ps->timeout = evSchedule(timeout,self,
			ps->timeout_time));
	xTrace0(photurisp, TR_GROSS_EVENTS, "photuris_init finished");

}

static xkern_return_t
photurisClose( self )
    XObj      self;
{
      SState * ss = (SState *) self->state; 
      XObj lls;
      lls=xGetDown(self,0);
      xTrace0(photurisp, TR_GROSS_EVENTS, "photurisClose entered");
      xControl(xGetDown(self,0),UDP_RELEASEPORTNUM,(char *)&ss->local_port,sizeof(long));
      if (ss->incomming_key_message != NULL)
              xFree((char *) ss->incomming_key_message);
      if (ss->outgoing_key_message != NULL)
              xFree((char *) ss->outgoing_key_message);
      if (ss->incomming_signature_message != NULL)
              xFree((char *) ss->incomming_signature_message);
      if (ss->outgoing_signature_message != NULL)
              xFree((char *) ss->outgoing_signature_message);
      if (ss->gx_session != NULL)
		xClose(ss->gx_session);
      xFree(self->state);
      mapUnbind( ((PState*)self->myprotl->state)->actMap,&lls);
      xClose(lls);
      xDestroy(self);
      xTrace0(photurisp, TR_GROSS_EVENTS, "photurisClose finished");
      return XK_SUCCESS;
}

static void
photurisSessnInit(s)
    XObj s;
{ 
  	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisSessnInit entered");
 	s->close = photurisClose;
  	s->pop = photurisPop;
 	xTrace0(photurisp, TR_GROSS_EVENTS, "photurisSessnInit done");
}
