/*
 * tcp_var.h
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_var.h	7.6 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:40:04 $
 */

/*
 * xtcp configuration:  This is a half-assed attempt to make xtcp
 * self-configure for a few varieties of 4.2 and 4.3-based unixes.
 * If you don't have a) a 4.3bsd vax or b) a 3.x Sun (x<6), check
 * this carefully (it's probably not right).  Please send me mail
 * if you run into configuration problems.
 *  - Van Jacobson (van@lbl-csam.arpa)
 */

#ifndef BSD
#define BSD 42	/* if we're not 4.3, pretend we're 4.2 */
#endif

#if sun||BSD<43
#define TCP_R_COMPAT_42	/* set if we have to interop w/4.2 systems */
#endif

#ifndef SB_MAX
#ifdef SB_MAXCOUNT
#define	SB_MAX SB_MAXCOUNT	/* Sun has to be a little bit different... */
#else
#define SB_MAX 65535		/* XXX */
#endif /* SB_MAXCOUNT */
#endif /* SB_MAX */

/*
 * Bill Nowicki pointed out that the page size (CLBYTES) has
 * nothing to do with the mbuf cluster size.  So, we followed
 * Sun's lead and made the new define MCLBYTES stand for the mbuf
 * cluster size.  The following define makes up backwards compatible
 * with 4.3 and 4.2.  If CLBYTES is >1024 on your machine, check
 * this against the mbuf cluster definitions in /usr/include/sys/mbuf.h.
 */
#ifndef MCLBYTES
#define	MCLBYTES CLBYTES	/* XXX */
#endif

/*
 * The routine in_localaddr is broken in Sun's 3.4.  We redefine ours
 * (in tcp_r_input.c) so we use can it but won't have a name conflict.
 */
#ifdef sun
#define in_localaddr tcp_r_in_localaddr
#endif

/* --------------- end of xtcp_r config ---------------- */

/*
 * Kernel variables for tcp_r.
 */

struct reass {
  struct reass *next, *prev;
  struct tcp_rhdr th;
  Msg m;
};

/*
 * Tcp_R control block, one per tcp_r; fields:
 */
struct tcp_rcb {
	struct	reass *seg_next;	/* sequencing queue */
	struct	reass *seg_prev;
	short	t_state;		/* state of this connection */
	short	t_timer[TCP_RT_NTIMERS];	/* tcp_r timers */
	short	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	short	t_rxtcur;		/* current retransmit value */
	short	t_dupacks;		/* consecutive dup acks recd */
	u_short	t_maxseg;		/* maximum segment size */
	char	t_force;		/* 1 if forcing out a byte */
	u_char	t_flags;
#define	TF_ACKNOW	0x01		/* ack peer immediately */
#define	TF_DELACK	0x02		/* ack, but try to delay it */
#define	TF_NODELAY	0x04		/* don't delay packets to coalesce */
#define	TF_NOOPT	0x08		/* don't use tcp_r options */
#define	TF_SENTFIN	0x10		/* have sent FIN */
	struct	tcp_riphdr *t_template;	/* skeletal packet for transmit */
	struct	inpcb *t_inpcb;		/* back pointer to internet pcb */
/*
 * The following fields are used as in the protocol specification.
 * See RFC783, Dec. 1981, page 21.
 */
/* send sequence variables */
	tcp_r_seq	snd_una;		/* send unacknowledged */
	tcp_r_seq	snd_nxt;		/* send next */
	tcp_r_seq	snd_up;			/* send urgent pointer */
	tcp_r_seq	snd_wl1;		/* window update seg seq number */
	tcp_r_seq	snd_wl2;		/* window update seg ack number */
	tcp_r_seq	iss;			/* initial send sequence number */
	u_short	snd_wnd;		/* send window */
/* receive sequence variables */
	u_short	rcv_wnd;		/* receive window */
	tcp_r_seq	rcv_nxt;		/* receive next */
	tcp_r_seq	rcv_up;			/* receive urgent pointer */
	tcp_r_seq	irs;			/* initial receive sequence number */
/*
 * Additional variables for this implementation.
 */
/* receive variables */
	tcp_r_seq	rcv_adv;		/* advertised window */
/* retransmit variables */
	tcp_r_seq	snd_max;		/* highest sequence number sent
					 * used to recognize retransmits
					 */
/* congestion control (for slow start, source quench, retransmit after loss) */
	u_short	snd_cwnd;		/* congestion-controlled window */
	u_short snd_ssthresh;		/* snd_cwnd size threshhold for
					 * for slow start exponential to
					 * linear switch */
/*
 * transmit timing stuff.
 * srtt and rttvar are stored as fixed point; for convenience in smoothing,
 * srtt has 3 bits to the right of the binary point, rttvar has 2.
 * "Variance" is actually smoothed difference.
 */
	short	t_idle;			/* inactivity time */
	short	t_rtt;			/* round trip time */
	tcp_r_seq	t_rtseq;		/* sequence number being timed */
	short	t_srtt;			/* smoothed round-trip time */
	short	t_rttvar;		/* variance in round-trip time */
	u_short max_rcvd;		/* most peer has sent into window */
	u_short	max_sndwnd;		/* largest window peer has offered */
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */
#define	TCP_ROOB_HAVEDATA	0x01
#define	TCP_ROOB_HADDATA	0x02
};

#define	intotcp_rcb(ip)	((struct tcp_rcb *)(ip)->inp_ppcb)
#define	sototcp_rcb(so)	(intotcp_rcb(sotoinpcb(so)))

/*
 * TCP_R statistics.
 * Many of these should be kept per connection,
 * but that's inconvenient at the moment.
 */
struct	tcp_rstat {
#if BSD<=43
       int     tcp_rs_badsum;
       int     tcp_rs_badoff;
       int     tcp_rs_hdrops;
       int     tcp_rs_badsegs;
       int     tcp_rs_unack;
       /* 4.3+ BSD stats start here */
#endif
	u_long	tcp_rs_connattempt;	/* connections initiated */
	u_long	tcp_rs_accepts;		/* connections accepted */
	u_long	tcp_rs_connects;		/* connections established */
	u_long	tcp_rs_drops;		/* connections dropped */
	u_long	tcp_rs_conndrops;		/* embryonic connections dropped */
	u_long	tcp_rs_closed;		/* conn. closed (includes drops) */
	u_long	tcp_rs_segstimed;		/* segs where we tried to get rtt */
	u_long	tcp_rs_rttupdated;	/* times we succeeded */
	u_long	tcp_rs_delack;		/* delayed acks sent */
	u_long	tcp_rs_timeoutdrop;	/* conn. dropped in rxmt timeout */
	u_long	tcp_rs_rexmttimeo;	/* retransmit timeouts */
	u_long	tcp_rs_persisttimeo;	/* persist timeouts */
	u_long	tcp_rs_keeptimeo;		/* keepalive timeouts */
	u_long	tcp_rs_keepprobe;		/* keepalive probes sent */
	u_long	tcp_rs_keepdrops;		/* connections dropped in keepalive */

	u_long	tcp_rs_sndtotal;		/* total packets sent */
	u_long	tcp_rs_sndpack;		/* data packets sent */
	u_long	tcp_rs_sndbyte;		/* data bytes sent */
	u_long	tcp_rs_sndrexmitpack;	/* data packets retransmitted */
	u_long	tcp_rs_sndrexmitbyte;	/* data bytes retransmitted */
	u_long	tcp_rs_sndacks;		/* ack-only packets sent */
	u_long	tcp_rs_sndprobe;		/* window probes sent */
	u_long	tcp_rs_sndurg;		/* packets sent with URG only */
	u_long	tcp_rs_sndwinup;		/* window update-only packets sent */
	u_long	tcp_rs_sndctrl;		/* control (SYN|FIN|RST) packets sent */

	u_long	tcp_rs_rcvtotal;		/* total packets received */
	u_long	tcp_rs_rcvpack;		/* packets received in sequence */
	u_long	tcp_rs_rcvbyte;		/* bytes received in sequence */
	u_long	tcp_rs_rcvbadsum;		/* packets received with ccksum errs */
	u_long	tcp_rs_rcvbadoff;		/* packets received with bad offset */
	u_long	tcp_rs_rcvshort;		/* packets received too short */
	u_long	tcp_rs_rcvduppack;	/* duplicate-only packets received */
	u_long	tcp_rs_rcvdupbyte;	/* duplicate-only bytes received */
	u_long	tcp_rs_rcvpartduppack;	/* packets with some duplicate data */
	u_long	tcp_rs_rcvpartdupbyte;	/* dup. bytes in part-dup. packets */
	u_long	tcp_rs_rcvoopack;		/* out-of-order packets received */
	u_long	tcp_rs_rcvoobyte;		/* out-of-order bytes received */
	u_long	tcp_rs_rcvpackafterwin;	/* packets with data after window */
	u_long	tcp_rs_rcvbyteafterwin;	/* bytes rcvd after window */
	u_long	tcp_rs_rcvafterclose;	/* packets rcvd after "close" */
	u_long	tcp_rs_rcvwinprobe;	/* rcvd window probe packets */
	u_long	tcp_rs_rcvdupack;		/* rcvd duplicate acks */
	u_long	tcp_rs_rcvacktoomuch;	/* rcvd acks for unsent data */
	u_long	tcp_rs_rcvackpack;	/* rcvd ack packets */
	u_long	tcp_rs_rcvackbyte;	/* bytes acked by rcvd acks */
	u_long	tcp_rs_rcvwinupd;		/* rcvd window update packets */
};


#ifdef __STDC__

struct	tcp_riphdr *tcp_r_template( struct tcp_rcb * );
struct	tcp_rcb *tcp_r_destroy( struct tcp_rcb * );
struct  tcp_rcb *tcp_r_drop( struct tcp_rcb *, int );
struct	tcp_rcb *tcp_r_timers( struct tcp_rcb *, int );
struct  tcp_rcb *tcp_r_disconnect( struct tcp_rcb * );
struct  tcp_rcb *tcp_r_usrclosed( struct tcp_rcb * );

#else

struct	tcp_riphdr *tcp_r_template();
struct	tcp_rcb *tcp_r_destroy();
struct  tcp_rcb *tcp_r_drop();
struct	tcp_rcb *tcp_r_timers();
struct  tcp_rcb *tcp_r_disconnect();
struct  tcp_rcb *tcp_r_usrclosed();

#endif /* __STDC__ */
