%{
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#define STDPRTLPATH "/cs/x32/merge/protocols/"

char name[64], prot[64], protName1[64], instName1[64], protName2[64], instName2[64];
struct _namelist {
  int len;
  char **list;
} prots;

struct _kms {
  int len;
  char **list;
}  kms;

%}

BLANK	[ \t]*
IDENT	[a-zA-Z][a-zA-Z0-9_]*

%start  PROT NAME DIR ROMOPT

%%

^name{BLANK}={BLANK}		     { BEGIN NAME; }
^NAME{BLANK}={BLANK}		     { BEGIN NAME; }
<NAME>{IDENT}(\/{IDENT})?	     { int i; 
				       strcpy(name, yytext);
				       for(i = 0; (yytext[i] && yytext[i] != '/'); i++);
				       yytext[i] = 0;
				       strcpy(protName1, yytext);
				       printf("protocol(%s).\n", name);
				       
				       /* collect names of protocols to read in .pl files */
				       prots.list = (char **)realloc(prots.list, ++prots.len * sizeof(char *));
				       prots.list[prots.len-1] = (char *)malloc(strlen(protName1) + 1);
				       strcpy(prots.list[prots.len-1], protName1);
				       
				       /* collect names of km's */
				       if (!strcmp(protName1, "km") || !strcmp(protName1, "KM"))
					 {
					   kms.list = (char **)realloc(kms.list, 
								       ++kms.len * sizeof(char *));
					   kms.list[kms.len-1] = (char *)malloc(strlen(name) + 1);
					   strcpy(kms.list[kms.len-1], name);
					 }
				       
				       if (yytext[i] == '/') strcpy(instName1, &yytext[i+1]);
				       else instName1[0] = '\0';
				       BEGIN 0;
				     }


protocols{BLANK}={BLANK}	     { BEGIN PROT; }
PROTOCOLS{BLANK}={BLANK}	     { BEGIN PROT; }
<PROT>{IDENT}(\/{IDENT})?/{BLANK}\,  { int i = 0; 
                                       strcpy(prot, yytext);
				       while(yytext[i] && yytext[i] != '/') i++;
				       strncpy(protName2, yytext, i);
				       printf("link(%s,%s).\n", name, prot);
				       if (yytext[i] == '/') strcpy(instName2, &yytext[i+1]);
				       else instName2[0] = '\0';
				     }
<PROT>{IDENT}(\/{IDENT})?	     { int i = 0; 
				       strcpy(prot, yytext);
				       while(yytext[i] && yytext[i] != '/') i++;
				       strncpy(protName2, yytext, i);
				       printf("link(%s,%s).\n", name, prot);
				       if (yytext[i] == '/') strcpy(instName2, &yytext[i+1]);
				       else instName2[0] = '\0';
				       BEGIN 0;
				     }


dir{BLANK}={BLANK}                   { BEGIN DIR; }
DIR{BLANK}={BLANK}                   { BEGIN DIR; }
<DIR>[^ \t\n;]+                      { char *p;
                                       p = (char *)malloc(strlen(yytext) 
							  + strlen(prots.list[prots.len-1]) 
							  + 1);
				       strcpy(p, yytext);
				       strcat(p, "/");
				       strcat(p, prots.list[prots.len-1]);
				       free(prots.list[prots.len-1]);
				       prots.list[prots.len-1] = p;
				       BEGIN 0;
                                     }


^romopt{BLANK}                       { BEGIN ROMOPT; printf("romopt(["); }
<ROMOPT>[^ \t\n;]+                   { printf("%s,", yytext); }
<ROMOPT>[^ \t\n;]+/{BLANK};          { printf("%s]).\n", yytext); BEGIN 0; }
                          
^#[^\n]*			     ;
.				     ;
\n				     ;

%%

main(int argc, char **argv)
{
  int i, fd, fd0, fd1;
  char c, filename[128];

  prots.len = 0; prots.list = (char **)0;
  kms.len = 0;   kms.list = (char **)0;
  fd0 = dup(0); fd1 = dup(1);

  printf("/* facts protocol(P) and link(X,Y) generated by scan from graph.comp file */\n");
  yylex();

  if (!kms.len) printf("kms([]).\n");
  else 
    { 
      printf("kms([");
      for(i = 0; i < kms.len-1; i++)
	{
	  printf("%s,", kms.list[i]);
	}
      printf("%s]).\n", kms.list[i]);
    }

  
  printf("\n\n\n/* attributes of protocols read in from attribute files of each protocol */\n");
  for(i = 0; i < prots.len; i++)
    {
      sprintf(filename, "%s.pl", prots.list[i]); /* search the current dir first */
      if ((fd = open(filename, O_RDONLY)) == -1)
	{
	  sprintf(filename, "%s%s/%s.pl", STDPRTLPATH, prots.list[i], prots.list[i]);
	  fd = open(filename, O_RDONLY);
	}
      if (fd != -1)
	{
	  dup2(fd, 0);
	  while((c=getchar()) != -1) putchar(c);
	  dup2(fd0, 0);
	  close(fd);
	}
    }

  printf("\n\n\n/* main.pl */\n");
  fd = open("main.pl", O_RDONLY);
  if (fd != -1)
    {
      dup2(fd, 0);
      while((c=getchar()) != -1) putchar(c);
      dup2(fd0, 0);
      close(fd);
    }

  for(i = 1; i < argc; i++)
    {
      fd = open(argv[i], O_RDONLY);
      if (fd != -1)
	{
	  printf("\n\n\n/* %s */\n", argv[i]);
	  dup2(fd, 0);
	  while((c=getchar()) != -1) putchar(c);
	  dup2(fd0, 0);
	  close(fd);
	}
    }

  exit(0);
  /*  
  printf("\n?- ['main.pl'].\n");
  while(--argc) 
    {
      printf("?- ['%s'].\n", argv[argc]);
    }
  */
}
